function createModifierFunction(
	fn, modifiers, modifier_list
) {
	// this effectively makes a copy of the function
	const wrapped_fn = (...args) => fn(modifier_list, ...args);

	for (const modifier of modifiers) {
		Object.defineProperty(wrapped_fn, modifier, {
			get() {
				let new_modifiers = [...modifier_list];

				if (!new_modifiers.includes(modifier)) {
					new_modifiers.push(modifier);
				}

				return createModifierFunction(
					fn, modifiers, new_modifiers
				)
			}
		});
	}

	return wrapped_fn
}

function index(fn_impl, modifiers) {
	return createModifierFunction(fn_impl, modifiers, [])
}

function isNode() {
	if (typeof window !== "undefined") {
		return false
	}

	if (typeof process === "undefined") {
		return false
	}

	return true
}

var stylesMap = {
	//format: [ code, ansi escape code ]
	"black"     : [ 0, 30],
	"red"       : [ 1, 31],
	"green"     : [ 2, 32],
	"yellow"    : [ 3, 33],
	"blue"      : [ 4, 34],
	"magenta"   : [ 5, 35],
	"cyan"      : [ 6, 36],
	"white"     : [ 7, 37],
	"gray"      : [ 8, 90],

	"bgBlack"   : [10, 40],
	"bgRed"     : [11, 41],
	"bgGreen"   : [12, 42],
	"bgYellow"  : [13, 43],
	"bgBlue"    : [14, 44],
	"bgMagenta" : [15, 45],
	"bgCyan"    : [16, 46],
	"bgWhite"   : [17, 47],

	"bold"      : [20, 1],
	"italic"    : [21, 3],
	"underline" : [22, 4]
};

function applyStylesToString(styles, str) {
	let ret = ``;

	for (const style of styles) {
		const [code] = stylesMap[style];

		ret += `\u000b`;
		ret += String.fromCharCode(code);
	}

	ret += str;
	ret += `\u001b`.repeat(styles.length);

	return ret
}

function applyStyle(styles, ...args) {
	let ret = ``;

	for (const arg of args) {
		ret += applyStylesToString(styles, arg);
	}

	return ret
}

function createBuffer(input) {
	let instance = {
		position: 0,

		peek(n_bytes) {
			return input.slice(
				instance.position ,
				instance.position + n_bytes
			)
		},

		pop(n_bytes) {
			let tmp = instance.peek(n_bytes);

			instance.position += n_bytes;

			return tmp
		},

		end() {
			return instance.position >= input.length
		}
	};

	return instance
}

function styleCodeToName(style_code) {
	for (const style in stylesMap) {
		const [code] = stylesMap[style];

		if (style_code === String.fromCharCode(code)) {
			return style
		}
	}

	throw new Error(`No matching style name for code.`)
}

function parse$1(input) {
	let buffer = createBuffer(input);
	let ret = [];

	while (!buffer.end()) {
		const ch = buffer.pop(1);

		// read style
		if (ch === `\u000b`) {
			let style_code = buffer.pop(1);

			if (!style_code.length) break

			ret.push(["push", styleCodeToName(style_code)]);
		}
		// pop stack
		else if (ch === `\u001b`) {
			ret.push(["pop"]);
		}
		else {
			ret.push(ch);
		}
	}

	return ret
}

function styleNameToAnsiEscapeSequence(style_name, use_colors) {
	if (!use_colors) return ""

	for (const style in stylesMap) {
		const [code, ansi_escape_sequence] = stylesMap[style];

		if (style_name === style) {
			return `\u001b[${ansi_escape_sequence}m`
		}
	}
}

function render$1(tree, use_colors = true) {
	let stack = [];
	let ret = ``;

	for (const entry of tree) {
		if (typeof entry === "string") {
			ret += entry;
		}
		// push to stack
		else if (entry[0] === "push") {
			const style = entry[1];

			stack.push(style);

			ret += styleNameToAnsiEscapeSequence(style, use_colors);
		}
		// pop stack
		else if (entry[0] === "pop") {
			stack.pop();

			if (use_colors) ret += `\u001b[0;0m`;

			for (const style of stack) {
				ret += styleNameToAnsiEscapeSequence(style, use_colors);
			}
		}
	}

	return ret
}

function print_impl(args, printer, use_colors = true) {
	for (const arg of args) {
		const tree = parse$1(arg);
		const str = render$1(tree, use_colors);

		printer(str);
	}
}

const colorize = function(str) { return str; };

const styles = Object.keys(stylesMap);

for (const style in stylesMap) {
	colorize[style] = index((modifiers, ...args) => {
		return applyStyle([
			...modifiers,
			style
		], ...args)
	}, styles);
}

const c      = colorize;
const render = render$1;
const parse  = parse$1;

let node_modules = {};

if (isNode()) {
	node_modules.process = (await import('node:process')).default;
	node_modules.tty = (await import('node:tty')).default;
}

const print = (...args) => {
	if (!isNode()) {
		print_impl(args, (str) => console.log(str), true);

		return
	}

	const {tty, process} = node_modules;

	let use_colors = false;

	if (tty.isatty(process.stdout.fd)) {
		use_colors = true;
	}

	if ("FORCE_COLORS" in process.env) {
		use_colors = true;
	}

	print_impl(args, (str) => process.stdout.write(str), use_colors);
};

print.stderr = (...args) => {
	if (!isNode()) {
		print_impl(args, (str) => console.log(str), true);

		return
	}

	const {tty, process} = node_modules;

	let use_colors = false;

	if (tty.isatty(process.stderr.fd)) {
		use_colors = true;
	}

	if ("FORCE_COLORS" in process.env) {
		use_colors = true;
	}

	print_impl(args, (str) => process.stderr.write(str), use_colors);
};

export { c, parse, print, render };
